import pandas
import platform

def json_frame(df, id_prefixes=[], separator=''):
    return df.apply(lambda x: json_series(x, id_prefixes, separator), axis=1)

def extract_value_recursively(s, col_name, orig_col_name, data):
    parts = col_name.split('.')
    if len(parts) == 1:
        data[col_name] = s[orig_col_name]
        # now convert any NaNs scalars to nulls
        try:
            if not isinstance(data[col_name], list) and pandas.isnull(data[col_name]):
                data[col_name] = None
        except ValueError as vex:
            print('unable to extract column {}'.format(col_name))
            raise vex
        #print('val for {} is type {}'.format(orig_col_name, s[orig_col_name].__class__.__name__))
    else:
        data[parts[0]] = extract_value_recursively(s, '.'.join(parts[1:]), orig_col_name, data[parts[0]] \
                                                   if parts[0] in data else {})
    return data

def json_series(s, id_prefixes=[], separator=''):
    data = {}
    
    # now convert all values into a deep dict
    for col_name in s.index.values:
        val = extract_value_recursively(s, col_name, col_name, data)
        
    # add meta to root object
    if 'meta' not in data:
        data['meta'] = {}
    if 'meta' not in data['source']:
        data['source']['meta'] = {}
        
    # track the cluster node used to process this
    data['meta']['@node'] = platform.node()
    data['source']['meta']['row'] = s.name
    
    return data


import re 
import pandas
import jellyfish
from datetime import datetime

def common(series):
    return series.str.replace(r'\s+deceased', '', flags=re.IGNORECASE) \
        .str.replace(r'^\.+', '', regex=True) \
        .str.strip() \
        .str.title()

def id_common(series):
    return series.str.upper() \
        .str.replace("'", '') \
        .str.replace(r'\(((Flat|Unit).*)\)', r'\1', flags=re.IGNORECASE, regex=True) \
        .str.replace(r'^(\d{1,5}?\s*)\((.*)\)', r'\1\2', flags=re.IGNORECASE, regex=True) \
        .str.replace(r'^\((.*)\)', r'\1', regex=True) \
        .str.replace(r'\(.*\)', '', regex=True) \
        .str.replace(r'[^A-Z\d\s@•]+', '', regex=True) \
        .str.strip()

def forename(series):
    return common(series) \
        .str.replace(r'\s+[A-Z]$', '', regex=True, flags=re.IGNORECASE) \
        .map(lambda s: '?' if pandas.isna(s) or s == '' else s)

def surname(series):
    return common(series) \
        .map(lambda s: '?' if pandas.isna(s) or s == '' else s)

def company_name(series):
    return common(series)

def school_name(series):
    return common(series.fillna('Unlisted'))

def build_soundex(str_val, first_word):
    try:
        if str_val is None:
            return None
        elif first_word:
            return jellyfish.soundex(str_val.strip().split(' ')[0])
        return jellyfish.soundex(str_val)
    except Exception as ex:
        raise Exception('unable to build soundex for str_val "{}"'.format(str_val))

def resident_id(forename, surname, address_id):
    fsoundex = forename.map(lambda s: build_soundex(s, True))
    ssoundex = surname.map(lambda s: build_soundex(s, False))
    series = fsoundex + '•' + ssoundex + '@' + address_id
    return id_common(series) \
        .str.replace(r'\s+', ' ', regex=True) # always last move

def address_id(series):
    return id_common(series) \
        .str.replace(r'\s+', ' ', regex=True) # always last move

def company_id(series):
    return id_common(series.str.upper() \
        .str.replace(r'^.*\s+T/A\s+(.+)$', r'\1', regex=True) \
        .str.replace('LTD', '') \
        .str.replace('LIMITED', '') \
        .str.replace('LLP', '') \
        .str.replace('A/C', '') \
        .str.replace('&', ' AND ') \
        .str.replace(r'- REF:.*', '', regex=True) \
        .str.replace(r'^@', '', regex=True)) \
        .str.replace(r'\s+', ' ', regex=True) # always last move

def discount_code(series):
    return series.map(lambda s: None if pandas.isna(s) else discount_code_string(s.upper()))

def discount_code_string(code):
    if code is None:
        return code
    code = code.upper()
    if code in ['SINGLE', 'SPD'] or code.find('SINGLE') != -1:
        return 'SPD'
    if code.find('EMPTY') != -1 or code.find('UNINHAB') != -1 or code in ['NO RESIDENTS']:
        return 'EMPTY'
    # southend's raw empty codes
    if code.find('PCLC') == 0 or code.find('PCLD') == 0 or code in ['PREM100', 'PREM200', 'PREM300']:
        return 'EMPTY'
    return code

def parse_dates(s, date_formats):
    # try the formats we know about
    for date_format in date_formats:
        try:
            return datetime.strptime(s, date_format)
        except ValueError as ex:
            continue
    # last thing, try an iso parse
    try:
        return datetime.fromisoformat(s.split('.')[0])
    except ValueError as ex:
        raise Exception('unable to parse "{}" with formats: {} (and iso failed too)'.format(s, date_formats))

def convert_residents_to_companies(df, rule_name, func_map_name):
    companies_df = df.copy()
    companies_df['meta.original_lookup_id'] = companies_df['lookup_id'].copy()
    companies_df['meta.exceptional_rule_name'] = rule_name
    companies_df['company_name'] = company_name(func_map_name(companies_df))
    companies_df['lookup_id'] = company_id(companies_df['company_name'])
    companies_df['company_name'] = companies_df['company_name'].map(lambda s: s.split(' T/A ') if s.find(' T/A ') != -1 else s)
    companies_df['sbrr.active'] = False
    companies_df['sbrr.amount'] = 0
    companies_df.drop(columns=['forename', 'surname'], inplace=True)
    return companies_df


# test = pandas.DataFrame([
#     pandas.Series({'company_name': '4 Stars Private Limited T/A Red Chilliezs'})
# ])
# res = company_id(test['company_name'])
# print('res = {}'.format(res.iloc[0]))



# ability to import module from relative path
import sys; sys.path.insert(0,'./src'); sys.path.insert(0,'../src')

# determine 1st line of the address
from vipro_python.ai.address import simple_record

def expect(the_test, expected):
  got = simple_record(the_test)
  if got != expected:
    print('expected={}, got={}'.format(expected, got))
  assert got == expected

# TODO: get working
# def test_broken():
#   expect(['36', 'The Pinnacle', 'Victoria Avenue'],
#     '36 The Pinnacle Victoria Avenue')

#   expect(['Flat 2', '34', 'Elderton Road'],
#     'Flat 2, 34 Elderton Road')

#   expect(['43', '12', 'Fairfax Drive'],
#     'Flat 43, 12 Fairfax Drive')

#   expect(['23 FLAT', 'The Plaza', 'Royal Mews'],
#     'Flat 23, The Plaza Royal Mews')

#   expect(['41', '9 Chartwell Plaza', 'Southchurch Road'],
#     'Flat 41 9 Chartwell Plaza Southchurch Road')

#   expect(['22b', '15', 'Ashburnham Road'],
#     'Flat 22b 15 Ashburnham Road')

#   expect(['22', '15a', 'Ashburnham Road'],
#     'Flat 22 15a Ashburnham Road')

#   expect(['(Flat 10)', '15a', 'Ashburnham Road'],
#     'Flat 10 15a Ashburnham Road')

def test_address2():
  expect(['1 Glendaurel Court', 'Milton Road', 'SS0 7JU'],
    '1 Glendaurel Court Milton Road')

  expect(['1', 'Glendaurel Court', 'Milton Road', 'SS0 7JU'],
    '1 Glendaurel Court Milton Road')

  expect(['Suite 20a', 'Thamesgate House', '33/41', 'Victoria Avenue'],
    'Suite 20a Thamesgate House 33/41 Victoria Avenue')

  expect(['(Lower) 104 Shaftesbury Avenue', 'Essex', 'TT1 7LP'],
    '104 Shaftesbury Avenue')

  expect(['1 (Upper) Princes Court 25C Princes Street', 'Southend-On-Sea', 'Essex', 'Ss1 1Qa'],
    '1 Upper Princes Court 25C Princes Street')

  expect(['Southend Ymca (Flat 9 F/F) Newlands', '85 Ambleside Drive', 'Southend-On-Sea', 'Ss1 2Fy'],
    'Southend Ymca Flat 9 F/F Newlands, 85 Ambleside Drive')

  expect(['(Unit 7) 495A London Road', 'Westcliff-On-Sea', 'Ss0 9Lg'],
    'Unit 7 495A London Road')

  expect(['(FIRST FLOOR)', '19 ROYAL TERRACE', 'SOUTHEND-ON-SEA'],
    'FIRST FLOOR 19 ROYAL TERRACE')


from .address import frame

# extract addresses
def extract_address(df, full_address_field):
    return df.apply(lambda x: extract_address_series(x, full_address_field), axis=1, result_type='expand')

def extract_address_series(s, full_address_field):
    addrSplit = s[full_address_field].split(', ')
    for i in range(0, len(addrSplit)):
        s['Addr{}'.format(i+1)] = addrSplit[i]
        if i == len(addrSplit)-1:
            s['AddrLast'] = addrSplit[i]
    return s

def apply(df, full_address_field):
    # slow operation for large datasets
    print('WARNING: exploding addresses can take a long time with {} records'.format(len(df)))
    extracted_df = extract_address(df, full_address_field)
    addr_df = extracted_df.copy()
    addr_df['addr1_clean'] = frame(addr_df, ['Addr1', 'Addr2', 'Addr3', 'Addr4', 'Addr5', 'Addr6', 'Addr7', 'Addr8'])
    addr_df['AddrLast'] = addr_df['AddrLast'].str.upper()
    return addr_df
    
def cleanup(df):
    # which columns?
    colsToRm = ['AddrLast', 'addr1_clean']
    for i in range(0, 15):
        field = 'Addr{}'.format(i)
        if field in df:
            colsToRm.append(field)
    df.drop(columns=colsToRm, inplace=True)


import pandas
from nameparser import HumanName
import re

def frame(df, field):
    print('WARNING: can take time to explode the name for all {} rows'.format(len(df)))
    return df.apply(lambda x: series(x, field), axis=1)
    
def series(s, field):
    try:
        name = '?' if pandas.isna(s[field]) else s[field]
        name = re.sub('Mstr', 'Master', name, flags=re.IGNORECASE)
        parsed = HumanName(name)
        s['forename'] = parsed.first
        s['meta.middleName'] = None if parsed.middle == '' else parsed.middle
        s['surname'] = parsed.last
        return s
    except AttributeError as ex:
        raise Exception('unable to extract name in field {}, series {}'.format(field, s))


import re
import pandas

# potentially doesn't need to be exhaustive as it's only trying to catch possible outliers
street_identifiers = [
    'road', 'lane', 'avenue', 'passage', 'street', 'close', 'mews', 'parade', 'gardens', 'chase', 'court', 'esplanade', 'place', 'drive', 'grove',
]

def frame(df, addr_line_names):
    return df.apply(lambda x: series(x, addr_line_names), axis=1)

def series(series, addr_line_names):
    data = []
    for field in addr_line_names:
        if field in series:
            data.append(series[field])
    return simple_record(data)
    
def simple_record(data):
    ''' extract first line of the address '''
    try:
        data = ['' if pandas.isna(d) else d for d in data]
        result = data[0].strip()
        poss_road_field = 1
        is_short = lambda x: len(re.sub('flat', '', re.sub(r'\\s+', '', re.sub(',', '', x)), flags=re.IGNORECASE)) < 10
        # first line is all in brackets (southend ctd)
        if re.search(r'^\(.*\)$', result):
            result = re.sub(r'^\((.*)\)$', r'\1', result)
            if len(data) > 1:
                result += ' ' + data[1].strip()
        # two numbers without a flat prefix?
        if len(data) > 1 and re.search(r'^\d+(a|b|c|d|e)?$', result) and re.search(r'^\d+', data[1].strip()):
            result = 'Flat ' + result
        # separate flat number from street address with a comma
        if re.search('flat', result, re.IGNORECASE):
            result += ','
        # only number / flat number in 1st line?
        if is_short(result) and len(data) > 1:
            result += ' ' + data[1].strip()
            poss_road_field += 1
        # what about street name separated in addr2/3?
        if  len(data) > poss_road_field + 1 and (is_short(result) or any(re.search(si, data[poss_road_field], re.IGNORECASE) for si in street_identifiers)):
            result += ' ' + data[poss_road_field].strip()
        # what about street name separated in addr2 3 and 4?
        elif len(data) > poss_road_field + 1 and any(re.search(si, data[poss_road_field + 1], re.IGNORECASE) for si in street_identifiers):
            result += ' ' + data[poss_road_field].strip() + ' ' + data[poss_road_field + 1].strip()
        # flat prefix wrong way around?
        result = re.sub(r'([0-9]+)\s+flat', r'Flat \1', result, flags=re.IGNORECASE)
        result = re.sub(r'\(((Flat|Unit).*)\)', r'\1', result, flags=re.IGNORECASE)
        result = re.sub(r'^(\d{1,5}?\s*)\((.*)\)', r'\1\2', result, flags=re.IGNORECASE)
        result = re.sub(r'R\/O\s+', '', result, flags=re.IGNORECASE)
        result = re.sub(r'\(.*\)', '', result)
        return re.sub(r'\s+', ' ', result).strip()
    except AttributeError as e:
        print('error caused by input: {}'.format(data))
        raise e



import pandas
import re
import dateutil.parser

valid_authorities=['Basildon', 'Braintree', 'Brentwood', 'CastlePoint', 'Chelmsford', 'Colchester', 'EppingForest', 'EssexCounty', 'Harlow', 'Maldon', 'Rochford', 'Southend', 'Tendring', 'Thurrock', 'Uttlesford']
fields_global = ['@timestamp', 'authority', 'address_id', 'source.uri', 'full_address', 'lookup_id']
fields_children_schema = ['forename', 'surname', 'school.name']
fields_residents_schema = ['forename', 'surname']
fields_companies_schema = ['company_name', 'sbrr.active', 'sbrr.amount']
fields_children_dates = ['@timestamp', 'school.meta.entryDate']
always_for_surnames = ['NACRO', 'NOTTING HILL GENESIS', 'PASSMORES ACADEMY', 'SUPPORTWORKS', 'STREETS2HOMES', 'LIVABILITY']
company_words = ['GROUP', 'LTD', 'LIMITED', 'BOARD', 'SOCIETY', 'SOLUTIONS', 'SYSTEMS', 'CONSULTING', 'DEVELOPMENT', 'PROPERTIES', 'ESTATE', 'TRUST', 'PLC', 'COUNCIL', 'BRENTWOOD', 'CHELMSFORD', 'BASILDON', 'ASSOCIATION', 'HOUSING', 'MOSAIC', 'HARVEST', 'AND CARE', 'CARE HOMES', 'AGENCY', 'LLP', 'BOROUGH', '&', 'LETTINGS', 'SERVICE', 'HOLDINGS', 'WINE BAR', 'SOUTH EAST', 'UK OUTREACH', 'CORPORATION', 'CHARITABLE CORP', 'CENTRE', 'CHURCH', 'HOSPITAL', 'METALTECH', 'CARE NHS', 'TEMPLE OF', 'MAYOR', 'AUCTION SALE', 'ELECTRONICS', 'SOCIAL CLUB', 'BUTCHER', 'INTERNATIONAL', 'TRADING', 'HOMES', 'SCHOOL', 'NHS ', 'MANAGEMENT', 'GREENE KING', 'PARISH', 'CHARITIES', 'CHARITY', 'EVANGELIST', 'CO.', 'HOUSE ESSEX', 'ENGLAND', 'COMMUNITY', 'OF DEFENCE', 'INSTITUTION', 'PARTNERSHIP', 'NURSING HOME', 'FOUNDATION', 'RECREATION', 'CLUB', 'LIMTED', 'TEAM', 'MENTAL HEALTH', 'CHEMIST', 'PLACES', 'WORKS', 'OF THE', 'HOTEL', 'MEMORIAL', 'COMPANY', 'RESIDENTIAL']

def children(df):
    return df.apply(lambda x: series(x, fields_children_schema, fields_children_dates), axis=1)

def residents(df):
    return df.apply(lambda x: series(x, fields_residents_schema, fields_children_dates), axis=1)

def companies(df):
    return df.apply(lambda x: series(x, fields_companies_schema, fields_children_dates), axis=1)

def is_company(series):
    return series.map(lambda x: False if pandas.isna(x) else contains_company_word(x))

def contains_company_word(name):
    uname = name.upper()
    if uname in always_for_surnames:
        return True
    for word in company_words:
        if ' {}'.format(word) in uname:
            #print('detected word {} in {}'.format(word, uname))
            return True
    return False

def series(series, fields_schema, fields_dates):
    series_for(series, fields_global, 'global')
    series_for(series, fields_schema, 'domain-specific')
    series_dates(series, fields_dates)

def series_for(series, fields, category):
    for field in fields:
        if field not in series:
            raise Exception('missing field {}, category={}'.format(field, category))
        if isinstance(series[field], list):
            continue
        if pandas.isna(series[field]):
            raise Exception('empty value in field {}, category={}, series={}'.format(field, category, series))
        if field == 'authority' and series[field] not in valid_authorities:
            raise Exception('authority "{}" was not recognised, category={}, series={}'.format(series[field], category, series))
        if field == 'surname':
            if re.search(r'\s+deceased', series[field], flags=re.IGNORECASE):
                raise Exception('{} contains deceased flag, category={}, series={}'.format(field, category, series))
        if field in ['lookup_id', 'address_id']:
            if re.search(r'^[@•]', series[field], flags=re.IGNORECASE):
                raise Exception('{} contains • or @ prefix, category={}, series={}'.format(field, category, series))
        if field in ['dob', 'meta.startDate', 'meta.endDate', 'ctd.startDate', 'ctd.endDate']:
            if not pandas.isna(series[field]):
                if series[field].find('T') == -1 or series[field].find('/') == -1 or series[field].find(':') == -1:
                    raise Exception('{} contains a datetime but not in iso format, category={}, series={}'.format(field, category, series))

def series_dates(series, date_fields):
    for field in date_fields:
        if field in series:
            val = series[field]
            if not pandas.isna(val):
                dateutil.parser.isoparse(val)


import re
from datetime import datetime

PATTERN_FOR_CHELMSFORD=r'^([\d\/]+)\s*-\s*([\d\/]+)$'

def frame(df, field, date_format, pattern):
    return df.apply(lambda x: series(x, field, date_format, pattern), axis=1, result_type='expand')
    
def series(s, field, date_format, pattern):
    val = s[field]
    if val is None: return s
    res = re.match(pattern, val)
    s['meta.startDate'] = datetime.strptime(res.group(1), date_format).isoformat()
    s['meta.endDate'] = datetime.strptime(res.group(2), date_format).isoformat()
    return s



# ability to import module from relative path
import sys; sys.path.insert(0,'./src'); sys.path.insert(0,'../src')

# determine 1st line of the address
from vipro_python.ai.validate import contains_company_word

def test_company():
  res = contains_company_word("Anchor Hanover Group ( Helen Court)")
  assert res == True

def test_person():
  res = contains_company_word("Tom Medhurst")
  assert res == False